﻿###########################################################
#
#               Třídění souborů slučováním
#
# Implementace třídění souborů slučováním (mergesort). Pro 
# zjednodušení předpokládáme, že tříděný soubor obsahuje 
# hodnoty typu int, a to čtyřbajtové. Pro přístup k souborům 
# je implementována pomocná třída Soubor. Případná změna typu 
# dat v souboru vyžaduje úpravu této třídy, nic více 
# (za podmínky, že na hodnoty uložené v souboru lze aplikovat 
# relační operátory).
#
# Vlastní třídění implementuje funkce přirozené_slučování(), 
# která očekává jméno (cestu) souboru, který je třeba seřadit.

from random import shuffle
import struct, enum, sys, os

###########################################################
# Výčtový typ popisující režim souboru
class Režim(enum.Enum):
    """Výčtový typ popisující režim souboru"""
    Zavřen = 0
    Čtení = 1
    Zápis = 2

###########################################################
# Třída Soubor
#
class Soubor:
    """Třída Soubor pro manipulaci se soubory. Umožňuje 
       otevřít binární soubor pro čtení nebo pro zápis, 
       zavřít a smazat. Umožňuje také zjistit, zda jsme 
       na konci souboru před tím, než se pokusíme číst 
       další položku, a zjistit hodnotu této položky 
       před čtením."""
    def __init__(self, jméno: str, režim: Režim = Režim.Zavřen):
        """Konstruktor nastaví hodnoty datových složek a 
           je-li soubor vytvářen v režimu čtení, přečte první 
           položku a uloží ji do složky _další"""
        self._režim = režim
        self._jméno = jméno
        self._soubor = None
        self._další = None
        self._konec: bool = False
        if režim != Režim.Zavřen:
            self.otevři(režim)

    def otevři(self, režim: Režim):
        """Otevře soubor v zadaném režimu a nastaví jeho příznak. 
           Je-li soubir otevírán pro čtení, 
           přečte první položku a uloží ji do složky _další"""
        if režim == Režim.Zavřen: raise ValueError("Špatný režim otevření souboru")
        if self._soubor and (self._režim != Režim.Zavřen):
            self._soubor.close()
        self._soubor = open(self._jméno, "rb" if režim == Režim.Čtení else "wb")
        if not self._soubor: raise ValueError("Nepodařilo se otevřít soubor")
        self._režim = režim
        if režim == Režim.Čtení:
            x = self._soubor.read(4)
            self._konec = len(x) < 4
            if self.konec: return    # Když je soubor prázdný
            self._další, = struct.unpack("i", x)

    def zavři(self):
        """Zavře soubor a nastaví příznak režimu"""
        if self._soubor and (self._režim != Režim.Zavřen):
            self._soubor.close()
            self._režim = Režim.Zavřen

    def čti(self)->int:
        """Přečte další položku ze souboru"""
        if self._režim != Režim.Čtení or not self._soubor or self._konec:
            raise ValueError("Čtení v nesprávném režimu nebo za koncem souboru nebo soubor neexistuje")
        if not self.konec: výsledek = self._další
        x = self._soubor.read(4)
        self._konec = len(x) < 4
        if not self.konec:
            self._další, = struct.unpack("i", x)        
        else:
            self._další = None
        return výsledek

    def zapiš(self, co: int)->None:
        """Zapíše položku na konec souboru"""
        if self._režim != Režim.Zápis or not self._soubor:
            raise ValueError("Zápis v nesprávném režimu nebo soubor neexistuje")
        self._soubor.write(struct.pack("i", co))

    @property
    def příští(self):
        """Vrátí položku souboru, která je na řadě pro čtení"""
        return self._další

    @property
    def konec(self)->bool:
        """Vrátí hodnotu příznaku konce souboru"""
        return self._konec

    def smaž(self)->None:
        """Smaže zadaný soubor"""
        os.remove(self._jméno)

def vytvoř_soubor(jméno: str, velikost: int):
    """Vytvoří soubor, který bude tříděn. Soubor 
    obsahuje náhodně promíchaná čísla 0..velikost-1"""
    a = [i for i in range(0, velikost)]
    shuffle(a)
    s = Soubor(jméno, Režim.Zápis)
    for i in a:
        s.zapiš(i)
    s.zavři()

def vypiš_soubor(jméno: str):
    """Vypíše obsah zadaného souboru"""
    s = Soubor(jméno, Režim.Čtení)
    while not s.konec:
        print(s.čti(), end=" ")
    print()
    s.zavři()

def přirozené_slučování(jméno: str):
    """Implementace přirozeného slučování (mergesort)"""
    A = Soubor(jméno, Režim.Zavřen)
    B = Soubor("Data1.dta", Režim.Zavřen)
    C = Soubor("Data2.dta", Režim.Zavřen)
    počet_běhů: int = 0
    while True:
        B.otevři(Režim.Zápis)
        C.otevři(Režim.Zápis)
        A.otevři(Režim.Čtení)
        _rozděl(A, B, C)
        B.otevři(Režim.Čtení)
        C.otevři(Režim.Čtení)
        A.otevři(Režim.Zápis)
        počet_běhů = _sluč(A, B, C)
        if počet_běhů == 1: 
            A.zavři()
            B.zavři()
            B.smaž()
            C.zavři()
            C.smaž()
            break

def _rozděl(A: Soubor, B: Soubor, C: Soubor):
    """Rozdělí běhy ze souboru A do souborů B a C"""
    while True:
        _zkopíruj_běh(A, B)  # Z A do B
        if not A.konec:
            _zkopíruj_běh(A, C)
        if A.konec: break

def _zkopíruj_běh(X: Soubor, Y: Soubor):
    """Okopíruje jeden běh ze souboru X do Y"""
    konec_běhu: bool = False              # V cyklu kopíruje prvky,
    while True:                           # dokud nedojde na konec běhu
        konec_běhu = _zkopíruj_prvek(X, Y)
        if konec_běhu: break


def _zkopíruj_prvek(X: Soubor, Y: Soubor)->bool:
    """Okopíruje jeden prvek ze souboru X do Y"""
    pom = X.čti()
    Y.zapiš(pom)
    if X.konec: return True
    else: return pom > X.příští

def _sluč(A: Soubor, B: Soubor, C: Soubor) -> int:
    """Sloučí soubory B a C do A"""
    počet_běhů: int = 0
    while (not B.konec) and (not C.konec):
        _sluč_běh(A, B, C)
        počet_běhů += 1
    while not B.konec:
        _zkopíruj_běh(B, A)
        počet_běhů += 1
    while not C.konec:
        _zkopíruj_běh(C, A)
        počet_běhů += 1
    return počet_běhů

def _sluč_běh(A: Soubor, B: Soubor, C: Soubor):
    """Sloučí aktuální běh z B s aktuálním během z C a výsledek zapíše do A"""
    konec_běhu: bool = False
    while True:
        if B.příští < C.příští:
            konec_běhu = _zkopíruj_prvek(B, A)
            if konec_běhu: _zkopíruj_běh(C, A)
        else:
            konec_běhu = _zkopíruj_prvek(C, A)
            if konec_běhu: _zkopíruj_běh(B, A)
        if konec_běhu: break

def kontrola(jméno: str)->bool:
    """Vrátí logickou hodnotu říkající, zda je zadaný soubor seřazen"""
    s = Soubor(jméno, Režim.Čtení)
    pom1  = s.čti()
    while not s.konec:
        pom2  = s.čti()
        if pom1 > pom2: return False
        pom2 = pom1
    return True

def main()->int:
    velikost: int = 20                  # Velikost tříděného souboru
    jméno = "Pokus.dta"                 # Jméno tříděného souboru
    vytvoř_soubor(jméno, velikost)
    if velikost <= 50: vypiš_soubor(jméno)
    print(kontrola(jméno))
    přirozené_slučování(jméno)
    if velikost <= 50: vypiš_soubor(jméno)
    print(kontrola(jméno))
    return 0

if __name__ == "__main__":
    sys.exit(main())