﻿###########################################################
#
#                       AVL-strom 
# 
# Implementace AVL-stromu; program ukazuje vyvažovací 
# operace při vkládání do AVL-stromu a při odebírání z něj
#
# Implementace AVL-stromu je s úpravami převzata ze stránky
# https://www.geeksforgeeks.org/avl-tree-in-python/ 
# (citace z 14.6.2024)
# Metoda pro zobrazení stromu vznikla úpravou funkce
# pro zobrazení binárního stromu uvedené na
# https://stackoverflow.com/questions/34012886
# /print-binary-tree-level-by-level-in-python
###########################################################

class Vrchol:
    """Vrchol AVL-stromu"""
    def __init__(self, hodnota):
        """Konstruktor vytvoří vrchol se zadanou hodnotou"""
        self._hodnota = hodnota
        self._levý = None
        self._pravý = None
        self._výška = 1


class AVLStrom:
    """AVL-vyvážený binární vyhledávací strom"""
    def __init__(self):
        self._kořen = None

    #######################################################
    # Metody tvořící rozhraní třídy AVLStrom
    #######################################################
    def vlož(self, hodnota):
        """Vložení nové hodnoty do stromu"""
        self._kořen = self._vlož(self._kořen, hodnota)

    def odstraň(self, hodnota):
        """Odstranění zadané hodnoty ze stromu"""
        self._kořen = self._odstraň(self._kořen, hodnota)

    def najdi(self, hodnota):
        """Vyhledání vrcholu obsahujícího zadanou hodnotu. Vrátí odkaz na vrchol nebo 
        None, když tam zadaná hodnota není"""
        return self._najdi(self._kořen, hodnota)

    def zobraz(self):
        """Zobrazí binární strom v konzolovém okně"""
        řádky, *_ = self._připrav_zobrazení(self._kořen)
        for řádka in řádky:
            print(řádka)

    def výška(self):
        return self._výška(self._kořen)

    #######################################################
    # Pomocné metody sloužící k implementaci metod rozhraní 
    ####################################################### 
    def _výška(self, prvek):
        """Výška (počet pater) stromu počítáno od zadaného vrcholu"""
        if not prvek:
            return 0
        return prvek._výška

    def _vyváženost(self, prvek):
        """Zjistí vyváženost stromu v daném vrcholu"""
        if not prvek:
            return 0
        return self._výška(prvek._levý) - self._výška(prvek._pravý)

    def _vlož(self, kořen, hodnota):
        """Vložení nové hodnoty do podstromu určeného kořenem"""
        if not kořen:
            return Vrchol(hodnota)
        elif hodnota < kořen._hodnota:
            kořen._levý = self._vlož(kořen._levý, hodnota)
        else:
            kořen._pravý = self._vlož(kořen._pravý, hodnota)
        kořen.výška = 1 + max(self._výška(kořen._levý), self._výška(kořen._pravý))
        balance = self._vyváženost(kořen)
        # pravá rotace
        if balance > 1 and hodnota < kořen._levý._hodnota:
            return self._pravá_rotace(kořen)
        # levá rotace
        if balance < -1 and hodnota > kořen._pravý._hodnota:
            return self._levá_rotace(kořen)
        # levo-pravá rotace
        if balance > 1 and hodnota > kořen._levý._hodnota:
            kořen.levý = self._levá_rotace(kořen._levý)
            return self._pravá_rotace(kořen)
        # pravo-levá rotace
        if balance < -1 and hodnota < kořen._pravý._hodnota:
            kořen._pravý = self._pravá_rotace(kořen._pravý)
            return self._levá_rotace(kořen)
        return kořen

    def _odstraň(self, kořen, hodnota):
        """Odstranění zadané hodnoty z podstromu určeného kořenem"""
        if not kořen:
            return kořen
        if hodnota < kořen._hodnota:
            kořen._levý = self._odstraň(kořen._levý, hodnota)
        elif hodnota > kořen._hodnota:
            kořen._pravý = self._odstraň(kořen._pravý, hodnota)
        else:
            if not kořen._levý:
                temp = kořen._pravý
                kořen = None
                return temp
            elif not kořen._pravý:
                temp = kořen._levý
                kořen = None
                return temp
            temp = self._prvek_s_min_hodnotou(kořen._pravý)
            kořen._hodnota = temp._hodnota
            kořen._pravý = self._odstraň(kořen._pravý, temp._hodnota)
        if not kořen:
            return kořen
        kořen._výška = 1 + max(self._výška(kořen._levý), self._výška(kořen._pravý))
        balance = self._vyváženost(kořen)
        # levá rotace
        if balance > 1 and self._vyváženost(kořen._levý) >= 0:
            return self._pravá_rotace(kořen)
        # pravá rotace
        if balance < -1 and self.balance(kořen._pravý) <= 0:
            return self._levá_rotace(kořen)
        # levo-pravá rotace
        if balance > 1 and self.balance(kořen._levý) < 0:
            kořen.levý = self._levá_rotace(kořen.levý)
            return self._pravá_rotace(kořen)
        # pravo-levá rotace
        if balance < -1 and self.balance(kořen.pravý) > 0:
            kořen.pravý = self._pravá_rotace(kořen.pravý)
            return self._levá_rotace(kořen)
        return kořen

    def _levá_rotace(self, prvek):
        """Levá rotace vrcholů pro vyvážení stromu"""
        pom1 = prvek._pravý
        pom2 = pom1._levý
        pom1.levý = prvek
        prvek._pravý = pom2
        prvek._výška = 1 + max(self._výška(prvek._levý), self._výška(prvek._pravý))
        pom1._výška = 1 + max(self._výška(pom1._levý), self._výška(pom1._pravý))
        return pom1

    def _pravá_rotace(self, prvek):
        """Pravá rotace vrcholů pro vyvážení stromu"""
        pom3 = prvek._levý
        pom4 = pom3._pravý
        pom3._pravý = prvek
        prvek.levý = pom4
        prvek._výška = 1 + max(self._výška(prvek._levý), self._výška(prvek._pravý))
        pom3._výška = 1 + max(self._výška(pom3._levý), self._výška(pom3._pravý))
        return pom3

    def _prvek_s_min_hodnotou(self, kořen):
        """Vrátí odkaz na vrchol obsahující nejmenší hodnotu ve stromě"""
        aktuální = kořen
        while aktuální._levý:
            aktuální = aktuální._levý
        return aktuální

    def _najdi(self, kořen, hodnota):
        """Vrátí odkaz na vrchol obsahující zadanou hodnotu 
           nebo None, když tam zadaná hodnota není (pomocná 
           metoda)"""
        if not kořen or kořen._hodnota == hodnota:
            return kořen
        if kořen._hodnota < hodnota:
            return self._najdi(kořen._pravý, hodnota)
        return self._najdi(kořen._levý, hodnota)

    def _připrav_zobrazení(self, kořen: Vrchol):
        """Vrátí seznam řetězců, šířku, výšku a horizontální souřadnici kořene."""
        # Kořen bez potomků
        if kořen._pravý is None and kořen._levý is None:
            řádka = '%s' % kořen._hodnota
            šířka = len(řádka)
            výška = 1
            střed = šířka // 2
            return [řádka], šířka, výška, střed
        # Pouze levý potomek
        if kořen._pravý is None:
            řádky, n, p, x = self._připrav_zobrazení(kořen._levý)
            s = '%s' % kořen._hodnota
            u = len(s)
            první_řádka = (x + 1) * ' ' + (n - x - 1) * '_' + s
            druhá_řádka = x * ' ' + '/' + (n - x - 1 + u) * ' '
            posunuté_řádky = [line + u * ' ' for line in řádky]
            return [první_řádka, druhá_řádka] + posunuté_řádky, n + u, p + 2, n + u // 2
        # Pouze pravý potomek
        if kořen._levý is None:
            řádky, n, p, x = self._připrav_zobrazení(kořen._pravý)
            s = '%s' % kořen._hodnota
            u = len(s)
            první_řádka = s + x * '_' + (n - x) * ' '
            druhá_řádka = (u + x) * ' ' + '\\' + (n - x - 1) * ' '
            posunuté_řádky = [u * ' ' + řádka for řádka in řádky]
            return [první_řádka, druhá_řádka] + posunuté_řádky, n + u, p + 2, u // 2
        # Dva potomci
        levý, n, p, x = self._připrav_zobrazení(kořen._levý)
        pravý, m, q, y = self._připrav_zobrazení(kořen._pravý)
        s = '%s' % kořen._hodnota
        u = len(s)
        první_řádka = (x + 1) * ' ' + (n - x - 1) * '_' + s + y * '_' + (m - y) * ' '
        druhá_řádka = x * ' ' + '/' + (n - x - 1 + u + y) * ' ' + '\\' + (m - y - 1) * ' '
        if p < q:
            levý += [n * ' '] * (q - p)
        elif q < p:
            pravý += [m * ' '] * (p - q)
        sbalené_řádky = zip(levý, pravý)
        řádky = [první_řádka, druhá_řádka] + [a + u * ' ' + b for a, b in sbalené_řádky]
        return řádky, n + m + u, max(p, q) + 2, n + u // 2

# Příklad chování AVL-stromu
def main()->int:
    strom = AVLStrom()
    strom.vlož(4)
    strom.zobraz()
    print("---------------")
    strom.vlož(5)
    strom.vlož(7)
    strom.zobraz()
    print("---------------")
    strom.vlož(2)
    strom.vlož(1)
    strom.zobraz()
    print("---------------")
    strom.vlož(3)
    strom.zobraz()
    print("---------------")
    strom.vlož(6)
    strom.zobraz()
    print("---------------")
    strom.odstraň(4)
    strom.zobraz()
    print("---------------")
    print(f"Výška stromu: {strom.výška()}")
    x = strom.najdi(5)
    print("Nalezeno" if x else "Nenalezeno")
    return 0

if __name__ == "__main__":
    exit(main())

    