﻿###########################################################
#
#               Kalkulačka založená na RPZ
#
# Program poskytuje kalkulačku pracující v příkazové řádce;
# výpočet je založen na obráceném polském zápisu. Proměnnné 
# se definují výskytem. Jméno proměnné je libovolný řetězec 
# složený z písmen a číslic začínající písmenem, hodnotu 
# proměnné definujeme přiřazením pomocí operátoru =. Výrazy 
# mohou obsahovat binární operátory +, -, * a /, unární 
# operátory + a - a kulaté závorky. 
# Po spuštění program zná hodnoty konstant π a e (pí, e)
#
# Program ukončíme zadáním slova "konec" (na velikosti 
# písmen nezáleží). 
#
###########################################################
#
# Program nejprve převede zadaný znakový řetězec na 
# posloupnost symbolů (token) reprezentovaných pomocí instancí
# třídy "Symbol"; ten pak převede RPZ tvořeného posloupností 
# symbolů (pouze operandů - čísel a proměnných - a operátorů, 
# bez závorek) a ten vyhodnotí.
# Proměnné jsou v programu reprezentovány 
# slovníkem (dict) "proměnná"; klíčem zde je jméno proměnné, 
# hodnotou je hodnota uložená v proměnné.
#
# Ve srovnání s programem RPZ2 je zde navíc rozklad vstupu 
# na symboly (token). Do RPZ se pak převádí posloupnost 
# symbolů. Ovšem funkce čti_symbol() nerozliší unární 
# a binární operátory, proto je posloupnost symbolů dále 
# upravena funkcí unární_operátory(), která je rozliší. 
# Teprve pak lze posloupnost symbolů převést do RPZ 
# a vypočítat hodnotu výrazu.
###########################################################

from enum import Enum
from symtable import Symbol
import sys, math

global priorita, proměnná

#tabulka priorit operátorů; předpona 'u' znamená unární operátor
priorita: dict =  {"=": 1, "+": 2, "-": 2, "*": 3, "/": 3, "u+": 4, "u-": 4}
#tabulka proměnných
proměnná: dict = {"pi": math.pi, "e": math.e}

class TypSymbolu(Enum):
    """Výčtový typ pro reprezentaci typu symbolu"""
    NEDEFINOVÁNO = 0
    LEVÁ_ZÁVORKA = 1
    PRAVÁ_ZÁVORKA = 2
    IDENTIFIKÁTOR = 3
    ČÍSLO = 4
    OPERÁTOR = 5
    KONEC = 6               # Reprezentuje poslední položku
    
class Symbol:
    """Třída představující symbol ve výrazu"""
    def __init__(self, _i = 0, _typ = TypSymbolu.NEDEFINOVÁNO, _hodnota = None) -> None:
        self.typ = _typ
        self.hodnota = _hodnota
        self.i = _i            # Pomocná hodnota pro usnadnění orientace ve vstupním řetězci
    #def __str__(self):
    #    return "{" + str(self.typ) + "; " + \
    #           str(self.hodnota) + "; " + str(self.i) + "}"

###########################################################
# Převedení vstupního řetězce na pole symbolů 
###########################################################
def čti_symbol(vstup: str, i: int)->Symbol:
    """Přečte ze vstupního řetězce další symbol počínaje i-tým znakem a vrátí ho"""
    if i == len(vstup): return Symbol(0, TypSymbolu.KONEC)
    while i < len(vstup) and vstup[i] == ' ':
        if i == len(vstup): return Symbol(0, TypSymbolu.KONEC)
        i += 1
    if i == len(vstup): return Symbol(0, TypSymbolu.KONEC)
    match vstup[i]:
        case '(':
            i += 1
            return Symbol(i, TypSymbolu.LEVÁ_ZÁVORKA, '(')
        case ')':
            i += 1
            return Symbol(i, TypSymbolu.PRAVÁ_ZÁVORKA, ')')
        case '+'|'-'|'*'|'/'|'=':
            i += 1
            return Symbol(i, TypSymbolu.OPERÁTOR, vstup[i-1])
        case '0'|'1'|'2'|'3'|'4'|'5'|'6'|'7'|'8'|'9'|'.':
            pom = ""
            while str(vstup[i]).isdigit() or vstup[i] == '.':
                pom = pom + vstup[i]
                i += 1
                if i == len(vstup): break
            hodnota = float(pom)
            return Symbol(i, TypSymbolu.ČÍSLO, hodnota)
        case _:
            if not str(vstup[i]).isalpha(): raise ValueError("Nedovolený symbol")
            pom = ""
            while i < len(vstup) and str(vstup[i]).isalnum():
                pom = pom + vstup[i]
                i += 1
            return Symbol(i, TypSymbolu.IDENTIFIKÁTOR, pom)

def vytvoř_pole_symbolů(vstup: str)->list:
    """Vytvoří ze vstupního řetězce pole symbolů (instancí třídy Symbol),
       nerozliší však unární a binární + a -"""
    symboly = []
    symboly.append(čti_symbol(vstup, 0))
    přepravka = symboly[0]
    while přepravka.typ != TypSymbolu.KONEC:
        přepravka = čti_symbol(vstup, přepravka.i)
        symboly.append(přepravka)
        if přepravka.typ == TypSymbolu.KONEC: break
    return symboly

def urči_unární_operátory(symboly: list)->None:
    """Identifikuje v řetězci symbolů unární operátory + a - a označí je"""
    if symboly[0].hodnota == '+' or symboly[0].hodnota == '-':
        symboly[0].hodnota = 'u' + symboly[0].hodnota
    for i in range(1, len(symboly)):
        if (symboly[i].hodnota == '-' or symboly[i].hodnota == '+'):
            if (symboly[i - 1].hodnota == '(' or symboly[i - 1].typ == TypSymbolu.OPERÁTOR):
                symboly[i].hodnota = 'u'+ symboly[i].hodnota

###########################################################
#  Převod pole symbolů do obráceného polského zápisu
###########################################################

def zpracuj_operátor(symbol: Symbol, zásobník: list, rpz: list):
    """Pomocná metoda pro zpracování operátoru při převodu do RPZ"""
    global priorita
    while True:
        if not zásobník: 
            zásobník.append(symbol)
            return
        else:
            pom = zásobník[-1]
            if pom.hodnota == "(" or priorita[symbol.hodnota] > priorita[pom.hodnota]:
                zásobník.append(symbol)
                return
            else:
                zásobník.pop()
                rpz.append(pom)

def zpracuj_pravou_závorku(zásobník: list, rpz: list):
    """Pomocná metoda pro zpracování pravé závorky při převodu do RPZ"""
    while True:
        if not zásobník: raise Exception("Nesouhlas závorek")
        pom = zásobník.pop()
        if pom.hodnota == "(": break
        rpz.append(pom)

def symboly_do_rpz(symboly:list)->list:
    """Převod pole symbolů do obráceného polského zápisu"""
    zásobník: list = []
    rpz: list = [] 
    for symbol in symboly:
        if symbol.hodnota == "(":
            zásobník.append(symbol)
            continue
        elif symbol.typ == TypSymbolu.IDENTIFIKÁTOR or symbol.typ == TypSymbolu.ČÍSLO:
            rpz.append(symbol)
            continue
        elif symbol.typ == TypSymbolu.OPERÁTOR:
            zpracuj_operátor(symbol, zásobník, rpz)
            continue
        elif symbol.hodnota == ")":
            zpracuj_pravou_závorku(zásobník, rpz)
            continue
        elif symbol.typ == TypSymbolu.KONEC: break
        raise Exception("Nedovolený znak")
    while zásobník:
        pom = zásobník.pop()
        if pom.hodnota == "(": raise Exception("Nesouhlas závorek")
        rpz.append(pom)
    return rpz

###########################################################
# Výpočet hodnoty výrazu zadaného vstupním řetězcem
###########################################################

def je_operand(symbol: Symbol)->bool:
    """Vrací True, je-li předaný symbol operandem"""
    return symbol.typ == TypSymbolu.IDENTIFIKÁTOR or symbol.typ == TypSymbolu.ČÍSLO

def je_unární(symbol: Symbol)-> bool:
    """Vrací True, je-li předaný symbol unárním operátorem"""
    return symbol.hodnota == "u+" or symbol.hodnota == "u-"

def hodnota_operandu(operand: Symbol)->float:
    """Vrátí hodnotu operandu reprezentovaného předaným symbolem"""
    if operand.typ == TypSymbolu.ČÍSLO: 
        return operand.hodnota
    else:
        if operand.hodnota not in proměnná:
            proměnná[operand.hodnota] = 0
        return proměnná[operand.hodnota]
    
def unární_operace(operátor: Symbol, operand: Symbol)->float:
    """Provede operaci určenou unárním operátorem"""
    pom = hodnota_operandu(operand)
    return pom if operátor.hodnota == "u+" else -pom

def binární_operace(operátor: Symbol, operand1: Symbol, operand2: Symbol)->float:
    """Provede operaci určenou binárním operátorem"""
    match operátor.hodnota:
        case "+": return hodnota_operandu(operand2) + hodnota_operandu(operand1)
        case "-": return hodnota_operandu(operand2) - hodnota_operandu(operand1)
        case "*": return hodnota_operandu(operand2) * hodnota_operandu(operand1)
        case "/": return hodnota_operandu(operand2) / hodnota_operandu(operand1)

def výpočet(rpz: list)-> float:
    """Výpočet hodnoty výrazu reprezentovaného obráceným polským zápisem 
       složeným z instancí třídy Symbol"""
    if not rpz:
        raise ValueError("Je třeba zadat výraz")  
    if len(rpz) == 1: return hodnota_operandu(rpz.pop())
    zásobník: list = []
    for i in range(0, len(rpz)):
        symbol = rpz[i]
        if je_operand(symbol):
            zásobník.append(symbol)
            continue
        else:
            if symbol.hodnota == "=":
                hodnota = zásobník.pop().hodnota
                identifikátor = zásobník.pop().hodnota
                proměnná[identifikátor] = hodnota
                return hodnota
            elif je_unární(symbol):
                hodnota = unární_operace(symbol, zásobník.pop())  
            else: 
                if len(zásobník) < 2: raise ValueError("Chybně zadaný výraz")
                hodnota = binární_operace(symbol, zásobník.pop(), zásobník.pop())
            zásobník.append(Symbol(0, TypSymbolu.ČÍSLO, hodnota)) 
    return zásobník.pop().hodnota
        
###########################################################
# Zpracování vstupního řetězce: Převod do RPZ a vyhodnocení
###########################################################
def zpracuj(vstupní_text: str)->float:
    symboly = vytvoř_pole_symbolů(vstupní_text)
    urči_unární_operátory(symboly)
    rpz = symboly_do_rpz(symboly)
    return výpočet(rpz)

def main()->int:
    while True:
        try:
            zadáno: str = input("Zadej výraz: ")
            if zadáno.lower() == "konec": break
            print(zpracuj(zadáno))
        except Exception as výjimka:
            print(výjimka)
    return 0

if __name__ == '__main__':
    sys.exit(main())





